/* -----------------------------------------------------------------------------

 datatype.api 1999 Dietmar Eilert

 Plug-in example. This plug-in show a file (image, guide etc.) in a container
 using the datatypes library.

 KNOWN BUGS:

 - The AmigaGuide datatype fails to determine its height and dimensions
   correctly sometimes (depending on its placement in the window). It seems
   to assume that it can use the full window and actually renders outside its
   box. You can minimize the problem by not placing the container on the left
   side of the window.

 - The AmigaGuide datatype creates gadgets with IDs outside the reserved
   namespace (add-ons may not choose gadget IDs freely because they must
   share the window with other add-ons). Editor gadgets will stop working
   for this reason.

 DICE:

 DMAKE

 -------------------------------------------------------------------------------

*/

#include "defs.h"

/// "prototypes"

// library functions

Prototype LibCall struct APIClient *APIMountClient(__A0 struct APIMessage *, __A1 char *);
Prototype LibCall void              APICloseClient(__A0 struct APIClient  *, __A1 struct APIMessage *);
Prototype LibCall void              APIBriefClient(__A0 struct APIClient  *, __A1 struct APIMessage *);
Prototype LibCall void              APIFree       (__A0 struct APIClient  *, __A1 struct APIOrder   *);

// private functions

Prototype ULONG                     AttachDTObject   (struct PlugInContext *, struct APIMessage *);
Prototype UWORD                     ComputeX         (struct TextFont *, UWORD);
Prototype UWORD                     ComputeY         (struct TextFont *, UWORD);
Prototype ULONG                     DetachDTObject   (struct PlugInContext *, struct APIMessage *);
Prototype ULONG                     DispatchContainer(struct APIClient     *, struct APIMessage *);
Prototype void                      GetVScreenSize   (struct Screen *, UWORD *, UWORD *);
Prototype struct Window            *Message          (UBYTE *, BOOL, struct Window *);
Prototype struct Window            *OpenMessageWin   (UBYTE *, BOOL);
Prototype BOOL                      SetSource        (struct PlugInContext *, UBYTE *);

///
/// "defines"

// we allocate a context structure for each instance of this client to store local data

struct PlugInContext {

    struct APIClient        APIClient;               // API handle (exposed to host)
    UBYTE                  *SourceFile;              // file to be displayed
    Object                 *Object;                  // datatype object
};

///
/// "library functions"

LibCall struct APIClient *
APIMountClient(__A0 struct APIMessage *apiMsg, __A1 char *args)
{
    struct PlugInContext *context;

    // build a description of this client (host will pass this back to us as "handle" during subsequent library calls)

    if (context = (struct PlugInContext *)AllocVec(sizeof(struct PlugInContext), MEMF_ANY | MEMF_CLEAR)) {

        if (SetSource(context, args)) {

            if (context->APIClient.api_Area = (struct APIArea *)AllocVec(sizeof(struct APIArea), MEMF_ANY | MEMF_CLEAR)) {

                // client description

                context->APIClient.api_APIVersion = API_INTERFACE_VERSION;
                context->APIClient.api_Version    = 1;
                context->APIClient.api_Name       = "Datatype";
                context->APIClient.api_Info       = "Plug-in example";
                context->APIClient.api_Commands   = NULL;
                context->APIClient.api_Serial     = 0;
                context->APIClient.api_Classes    = API_CLASS_SYSTEM | API_CLASS_CONTAINER;

                // prepare the container request

                context->APIClient.api_Area->api_Width     = 50;
                context->APIClient.api_Area->api_Height    = 10;
                context->APIClient.api_Area->api_UnitsX    = API_UNITS_PERCENT;
                context->APIClient.api_Area->api_UnitsY    = API_UNITS_FONT;
                context->APIClient.api_Area->api_Alignment = API_ALIGN_RIGHT;
                context->APIClient.api_Area->api_Style     = API_STYLE_STANDARD;
                context->APIClient.api_Area->api_IDCMP     = IDCMP_IDCMPUPDATE;
            }
            else {

                FreeVec(context);

                context = NULL;
            }
        }
        else {

            FreeVec(context);

            context = NULL;
        }
    }

    return((struct APIClient *)context);
}

LibCall void
APICloseClient(__A0 struct APIClient *handle, __A1 struct APIMessage *apiMsg)
{
    // free ressources related to this client

    FreeVec(handle->api_Area);

    FreeVec(handle);
}

LibCall void
APIBriefClient(__A0 struct APIClient *handle, __A1 struct APIMessage *apiMsg)
{
    struct APIMessage *msg;

    // handle host's command notify

    for (msg = apiMsg; msg; msg = msg->api_Next) {

        if (msg->api_State == API_STATE_NOTIFY) {

            switch (msg->api_Class) {

                case API_CLASS_CONTAINER:

                    msg->api_Error = DispatchContainer(handle, msg);

                    break;

                case API_CLASS_SYSTEM:

                    msg->api_Error = API_ERROR_OK;

                    break;

                default:

                    msg->api_Error = API_ERROR_UNKNOWN;
            }
        }
    }
}


LibCall void
APIFree(__A0 struct APIClient *handle, __A1 struct APIOrder *apiOrder)
{
    struct APIOrder *order, *next;

    // free data allocated for orders sent to host

    for (order = apiOrder; order; order = next) {

        next = order->api_Next;

        switch (order->api_Flags) {

            case API_ORDER_DTDETACH:

                DisposeDTObject((Object *)order->api_Data);

                break;
        }

        FreeVec(order);
    }
}

///
/// "gui"

/* ---------------------------------- Message ----------------------------------

 Show message (assume that window size is sufficient)

*/

struct Window *
Message(text, sync, win)

struct Window *win;
UBYTE         *text;
BOOL           sync;
{
    if (win == NULL)

        win = OpenMessageWin(text, sync);

    else {

        struct RastPort *rast = win->RPort;

        SetAPen(rast, 0);
        SetRast(rast, 0);
        SetAPen(rast, 1);

        Move(rast, 10, 10 + rast->TxBaseline);

        Text(rast, text, strlen(text));

        if (sync) {

            Delay(150);

            CloseWindow(win);

            win = NULL;
        }
    }

    return(win);
}


/* -------------------------------- OpenMessageWin -----------------------------

 Open info window. Window is closed automatically after a short delay time if
 <sync> is TRUE.

*/

struct Window *
OpenMessageWin(text, sync)

UBYTE *text;
BOOL   sync;
{
    struct TextFont *font;
    struct Screen   *screen;
    struct Window   *win;
    ULONG            lock;

    win  = NULL;
    font = NULL;

    // don't do this at home, kids ;-)

    lock = LockIBase(0);

    // find frontmost screen

    screen = ((struct IntuitionBase *)IntuitionBase)->ActiveScreen;

    // keep system frozen

    Forbid();

    // re-enable rendering functions

    UnlockIBase(lock);

    if (screen) {

        struct TextAttr textAttr = { "topaz.font", 8, 0, FPF_DESIGNED };

        WORD len, width, height, fontW, fontH, x, y, displayW, displayH, xOffset, yOffset;

        GetVScreenSize(screen, &displayW, &displayH);

        xOffset = -screen->ViewPort.DxOffset;
        yOffset = -screen->ViewPort.DyOffset;

        textAttr = *screen->Font;

        if (font = OpenDiskFont(&textAttr)) {

            fontW = font->tf_XSize;
            fontH = font->tf_YSize;

            CloseFont(font);
        }
        else
            fontW = fontH = textAttr.ta_YSize;

        len = strlen(text);

        if (len < 30)

            len = 30;

        width  = 20 + fontW * len;
        height = 20 + fontH;

        // center window on screen

        if (displayW) {

            x = ((displayW - width )>>1) + xOffset;
            y = ((displayH - height)>>1) + yOffset;
        }
        else
            x = y = 0;

        win = OpenWindowTags(NULL,

            WA_PubScreen,     screen,
            WA_Left,          x,
            WA_Top,           y,
            WA_InnerWidth,    width,
            WA_InnerHeight,   height,
            WA_Title,        "Datatype Plug-In",
            WA_ScreenTitle,  "Datatype Plug-In",
            WA_Flags,         WFLG_DRAGBAR | WFLG_DEPTHGADGET | WFLG_SMART_REFRESH | WFLG_ACTIVATE | WFLG_RMBTRAP,
            WA_GimmeZeroZero, TRUE,
            TAG_DONE
        );

        Permit();

        if (win) {

            if (font)
                SetFont(win->RPort, font);

            win = Message(text, sync, win);
        }
    }
    else
        Permit();

    return(win);
}


/* ------------------------------ GetVScreenSize -------------------------------

 Calculate visible screen rectangle

*/

void
GetVScreenSize(screen, displayW, displayH)

struct Screen *screen;
UWORD         *displayW;
UWORD         *displayH;
{
    struct Rectangle clip;

    // read screen size

    if (QueryOverscan(GetVPModeID(&screen->ViewPort), &clip, OSCAN_TEXT)) {

        *displayW = clip.MaxX - clip.MinX + 1;
        *displayH = clip.MaxY - clip.MinY + 1;
    }
    else {

        *displayW = screen->Width;
        *displayH = screen->Height;
    }
}


/* --------------------------------- ComputeX ----------------------------------

 Scale width <value> to match font

*/

UWORD
ComputeX(font, value)

struct TextFont *font;
UWORD            value;
{
    return((font->tf_XSize * value) / 8);
}


/* --------------------------------- ComputeY ----------------------------------

 Resize height <value> to match font

*/

UWORD
ComputeY(font, value)

struct TextFont *font;
UWORD            value;
{
    return((font->tf_YSize * value) / 8);
}

///
/// "container"

/* ----------------------------- DispatchContainer -----------------------------

 Handle container-related event

*/

ULONG
DispatchContainer(handle, apiMsg)

struct APIClient  *handle;
struct APIMessage *apiMsg;
{
    ULONG error = API_ERROR_OK;

    switch (apiMsg->api_Action) {

        case API_ACTION_DETACH:

            // request to detach gadgets from window

            error = DetachDTObject((struct PlugInContext *)handle, apiMsg);

            break;

        case API_ACTION_ATTACH:

            // request to attach gadgets

            error = AttachDTObject((struct PlugInContext *)handle, apiMsg);

            break;

        case API_ACTION_REFRESH:

            // refresh user interface (nothing to do here: gadgets are refreshed automatically)

            break;

        case API_ACTION_PAINT:

            // render the user interface

            {
                struct PlugInContext *context = (struct PlugInContext *)handle;

                if (context->Object)

                    RefreshDTObjects(context->Object, apiMsg->api_Instance->api_Window, NULL, NULL);
            }

            break;

        case API_ACTION_INPUTEVENT:

            // input event to be handled

            struct APIInputEvent *event;

            if (event = (struct APIInputEvent *)apiMsg->api_Data) {

                switch (event->api_IDCMPClass) {

                    case IDCMP_IDCMPUPDATE:

                        {
                            // is this event meant for us ?

                            if (event->api_IDCMPWindow == apiMsg->api_Instance->api_Window) {

                                struct TagItem *tag, *tstate;

                                // get a pointer to the attributes

                                tstate = (struct TagItem *)event->api_IDCMPAddress;

                                while (tag = NextTagItem(&tstate)) {

                                    ULONG data = tag->ti_Data;

                                    switch (tag->ti_Tag) {

                                        case DTA_ErrorString:

                                            if (data)

                                                Message((UBYTE *)data, TRUE, NULL);

                                            break;

                                        case DTA_Sync:

                                            {
                                                struct PlugInContext *context = (struct PlugInContext *)handle;

                                                if (context->Object)

                                                    RefreshDTObjects(context->Object, apiMsg->api_Instance->api_Window, NULL, NULL);
                                            }

                                            break;
                                    }
                                }

                                // no further processing of this input event required

                                apiMsg->api_State = API_STATE_CONSUMED;
                            }
                        }

                        break;
                }
            }

            break;

        default:

            error = API_ERROR_UNKNOWN;
    }

    return(error);
}


/* ------------------------------- DetachDTObject ------------------------------

 Detach datatype object from window

*/

ULONG
DetachDTObject(context, apiMsg)

struct PlugInContext *context;
struct APIMessage    *apiMsg;
{
    // datatype object attached to window ?

    if (context->Object) {

        // prepare order to send a command ("detach objects from the window") to the host

        if (apiMsg->api_Order = (struct APIOrder *)AllocVec(sizeof(struct APIOrder), MEMF_PUBLIC | MEMF_CLEAR)) {

            // create order for host to have gadgets detached

            apiMsg->api_Order->api_Flags = API_ORDER_DTDETACH;

            apiMsg->api_Order->api_Data  = context->Object;

            // object now assumed to be detached from window

            context->Object = NULL;

            // see APIFree() function for further processing (disposal of objects)
        }
    }

    return(API_ERROR_OK);
}


/* ------------------------------- AttachDTObject ------------------------------

 Attach datatype object to window

*/

ULONG
AttachDTObject(context, apiMsg)

struct PlugInContext *context;
struct APIMessage    *apiMsg;
{
    ULONG error;

    // already attached to window ?

    if (context->Object) {

        error = API_ERROR_OK;
    }
    else {

        struct APIContainer *container;

        error = API_ERROR_FAIL;

        if (container = apiMsg->api_Instance->api_Container) {

            // container not fully obscured ?

            if (container->api_Clipping) {

                UWORD x0;
                UWORD y0;
                UWORD x1;
                UWORD y1;
                UWORD w;
                UWORD h;

                // determine space available for gadgets

                x0 = container->api_Clipping->MinX;
                y0 = container->api_Clipping->MinY;
                x1 = container->api_Clipping->MaxX;
                y1 = container->api_Clipping->MaxY;

                // center gadget in container

                w = (x1 - x0 + 1);
                h = (y1 - y0 + 1);

                if ((w >= 8) && (h >= 8)) {

                    Object *object = NewDTObject(context->SourceFile,

                        DTA_SourceType,  DTST_FILE,
                        GA_Left,         x0,
                        GA_Top,          y0,
                        GA_Width,        w,
                        GA_Height,       h,
                        DTA_TextAttr,    container->api_TextAttr,
                        GA_ID,           container->api_Namespace,
                        GA_DrawInfo,     container->api_DrawInfo,
                        ICA_TARGET,      ICTARGET_IDCMP,
                        TAG_DONE
                    );

                    if (object) {

                        // prepare order to send a command ("add these objects") to the host

                        if (apiMsg->api_Order = (struct APIOrder *)AllocVec(sizeof(struct APIOrder), MEMF_PUBLIC | MEMF_CLEAR)) {

                            // create order for host to have DT objects added to the window

                            apiMsg->api_Order->api_Flags = API_ORDER_DTATTACH;

                            apiMsg->api_Order->api_Data  = object;

                            // object now attached to window

                            context->Object = object;

                            // set return code

                            error = API_ERROR_OK;
                        }
                        else
                            DisposeDTObject(object);
                    }
                    else {

                        static UBYTE error[256];

                        Fault(IoErr(), "Datatype error", error, sizeof(error));

                        Message(error, TRUE, NULL);
                    }
                }
            }
        }
    }

    return(error);
}


/* --------------------------------- SetSource ---------------------------------

 Set source file (store path in context). Return TRUE if successful.

*/

BOOL
SetSource(context, file)

struct PlugInContext *context;
UBYTE                *file;
{
    BPTR lock = NULL;

    if (context->SourceFile)

        FreeVec(context->SourceFile);

    if (file && *file) {

        if (context->SourceFile = (UBYTE *)AllocVec(strlen(file) + 1, MEMF_ANY | MEMF_PUBLIC)) {

            strcpy(context->SourceFile, file);

            if (lock = Lock(file, ACCESS_READ)) {

                UnLock(lock);
            }
            else
                Message("File not found error", TRUE, NULL);
        }
    }
    else
        Message("No file specified", TRUE, NULL);

    if (lock)
        return(TRUE);
    else
        return(FALSE);
}

///
