/* -----------------------------------------------------------------------------

 mirror.api 1999 Dietmar Eilert

 API plug-in. Highlights matching bracket if cursor is moved over a bracket.

 Dice:

 DMAKE

 -------------------------------------------------------------------------------

*/

#include "defs.h"

/// "prototypes"

// library functions

Prototype LibCall struct APIClient *APIMountClient(__A0 struct APIMessage *, __A1 char *);
Prototype LibCall void              APICloseClient(__A0 struct APIClient *,  __A1 struct APIMessage *);
Prototype LibCall void              APIBriefClient(__A0 struct APIClient *,  __A1 struct APIMessage *);
Prototype LibCall void              APIFree       (__A0 struct APIClient  *, __A1 struct APIOrder   *);

// private functions

Prototype void Dispatch         (struct APIMessage *);
Prototype BOOL FindMarkedBracket(struct EditConfig *);
Prototype BOOL NoUserBlock      (struct EditConfig *);
Prototype BOOL IsABracket       (UWORD);
Prototype BOOL MatchingBracket  (struct APIMessage *);

///
/// "library functions"

LibCall struct APIClient *
APIMountClient(__A0 struct APIMessage *apiMsg, __A1 char *args)
{
    static struct APIClient apiClient;

    apiClient.api_APIVersion = API_INTERFACE_VERSION;
    apiClient.api_Version    = 3;
    apiClient.api_Name       = "Mirror";
    apiClient.api_Info       = "API example";
    apiClient.api_Commands   = NULL;
    apiClient.api_Serial     = 0;
    apiClient.api_Classes    = API_CLASS_SYSTEM | API_CLASS_KEY;
    apiClient.api_Area       = NULL;

    return(&apiClient);
}

LibCall void
APICloseClient(__A0 struct APIClient *handle, __A1 struct APIMessage *apiMsg)
{
    // no ressources to be freed
}

LibCall void
APIBriefClient(__A0 struct APIClient *handle, __A1 struct APIMessage *apiMsg)
{
    struct APIMessage *msg;

    // handle host's command notify

    for (msg = apiMsg; msg; msg = msg->api_Next) {

        if (msg->api_State == API_STATE_NOTIFY) {

            switch (msg->api_Class) {

                case API_CLASS_KEY:

                    Dispatch(msg);

                case API_CLASS_SYSTEM:

                    break;

                default:

                    msg->api_Error = API_ERROR_UNKNOWN;
            }
        }
    }
}

LibCall void
APIFree(__A0 struct APIClient *handle, __A1 struct APIOrder *apiOrder)
{
    struct APIOrder *order, *next;

    // free data allocated for orders sent to host

    for (order = apiOrder; order; order = next) {

        next = order->api_Next;

        FreeVec(order);
    }
}

///
/// "private functions"

/* --------------------------------- Dispatch ----------------------------------

 Dispatch incoming API event

*/

void
Dispatch(apiMsg)

struct APIMessage *apiMsg;
{
    struct EditConfig *config = (struct EditConfig *)apiMsg->api_Instance->api_Environment;

    // can we use block marking for our purposes (no block marked) ?

    if (NoUserBlock(config)) {

        switch (apiMsg->api_Action) {

            case API_ACTION_MOUSEBUTTON:
            case API_ACTION_RAWKEY:
            case API_ACTION_VANILLAKEY:

                {
                    UWORD column;
                    BOOL  check;

                    // check character just typed by user; should be the character under the cursor or before the cursor

                    if (column = config->Column) {

                        check = IsABracket(config->CurrentBuffer[column]);

                        if (check == FALSE)

                            check = IsABracket(config->CurrentBuffer[--config->Column]);
                    }
                    else
                        check = IsABracket(config->CurrentBuffer[column]);

                    if (check) {

                        // highlight matching bracket

                        if (MatchingBracket(apiMsg) == FALSE) {

                            // no marching bracket found: remove highlighting

                            if (FindMarkedBracket(config))

                                apiMsg->api_Refresh |= API_REFRESH_NOMARKER;
                        }
                    }
                    else if (FindMarkedBracket(config))

                        apiMsg->api_Refresh |= API_REFRESH_NOMARKER;

                    // restore cursor position

                    config->Column = column;
                }

                break;
        }
    }
}


/* ------------------------------- MatchingBracket -----------------------------

 Find and highlight matching bracket. Ignore "\}" if not preceded by another "\"
 (TEX constant).

*/

BOOL
MatchingBracket(apiMsg)

struct APIMessage *apiMsg;
{
    struct EditConfig *config;
    UBYTE             *current;
    UBYTE             *known;
    ULONG              line;
    UWORD              len;
    UWORD              column;

    config  = (struct EditConfig *)apiMsg->api_Instance->api_Environment;

    current = config->CurrentBuffer;
    line    = config->Line;
    len     = config->CurrentLen;
    column  = config->Column;

    // determine bracket type under cursor

    for (known = "(){}[]"; *known; ++known) {

        if (current[column] == *known) {

            BOOL ignoreTeX = ((column > 1) && (current[column - 1] == 92) && (current[column - 2] != 92));

            // skip over bracket if this is a TeX-style "\}" sequence

            if (ignoreTeX == FALSE) {

                BOOL  success;
                BOOL  instring;
                UWORD count;
                UWORD distance;
                UWORD twin;
                WORD  level;
                WORD  step;

                success  = FALSE;

                // determine direction of search and character to look for

                switch (*known) {

                    case '(':
                    case '{':
                    case '[':

                        step = 1;

                        break;

                    default:

                        step = -1;
                }

                twin = *(known + step);

                // tip: modify the scan depth (default: 50 lines) to increase performance

                instring =  FALSE;
                count    =  0;
                level    = -1;

                for (distance = 0; distance <= 50; ++distance) {

                    while ((column >= 0) && (column < len)) {

                        UBYTE *next = current + column;

                        if ((*next == 34) || (*next == 39))

                            instring = !instring;

                        else if (!instring) {

                            if ((column > 0) && (*(next - 1) == 92)) {

                                ignoreTeX = (column < 2) || (*(next - 2) != 92);
                            }
                            else
                                ignoreTeX = FALSE;

                            if (ignoreTeX == FALSE) {

                                if (*next == twin) {

                                    if (level) {

                                        --level;
                                    }
                                    else {

                                        success = TRUE;

                                        break;
                                    }
                                }
                                else if (*next == *known) {

                                    ++level;
                                    ++count;
                                }
                            }
                        }

                        column += step;
                    }

                    if (success) {

                        BOOL adjacent = FALSE;

                        if (line == config->Line)

                            adjacent = (column == (config->Column + step));

                        if (adjacent) {

                            // no highlighting for brackets directly next to cursor (too confusing)

                            apiMsg->api_Refresh |= API_REFRESH_NOMARKER;
                        }
                        else if (GET_FOLD(config->TextNodes + line)) {

                            // block markers can not be used in fold headers

                            return(FALSE);
                        }
                        else {

                            // create "set block marker" order for the editor

                            struct Order {

                                struct APIOrder        APIOrder;
                                struct APIBlockRequest APIBlockRequest;

                            } *apiOrder;

                            if (apiOrder = (struct Order *)AllocVec(sizeof(struct Order), MEMF_PUBLIC | MEMF_ANY)) {

                                apiOrder->APIOrder.api_Flags = API_ORDER_BLOCK;

                                apiOrder->APIOrder.api_Data  = &apiOrder->APIBlockRequest;

                                apiOrder->APIOrder.api_Next  = NULL;

                                // set new block dimensions

                                apiOrder->APIBlockRequest.br_Marker     = BLOCKMODE_CHAR;
                                apiOrder->APIBlockRequest.br_BlockStartX = column;
                                apiOrder->APIBlockRequest.br_BlockEndX   = column;
                                apiOrder->APIBlockRequest.br_BlockStartY = line;
                                apiOrder->APIBlockRequest.br_BlockEndY   = line;

                                // send order to editor

                                apiMsg->api_Order = &apiOrder->APIOrder;
                            }

                            return(TRUE);
                        }
                    }
                    else {

                        line    += step;
                        instring = FALSE;

                        if ((line >= 0) && (line < config->Lines)) {

                            struct LineNode *lineNode = config->TextNodes + line;

                            current = lineNode->Text;
                            len     = lineNode->Len;

                            column = (step == -1) ? (len - 1) : 0;
                        }
                        else                         // no matching bracket
                            break;
                    }
                }
            }

            break;
        }
    }

    return(FALSE);
}


/* ------------------------------- NoUserBlock ---------------------------------

 Return TRUE if there is no user-defined block, i.e. either no block at all or a
 marked bracket only (probably highlighted by us).

*/

BOOL
NoUserBlock(config)

struct EditConfig *config;
{
    if ((config->Marker == BLOCKMODE_NONE) || FindMarkedBracket(config))
        return(TRUE);
    else
        return(FALSE);

}


/* ----------------------------- FindMarkedBracket -----------------------------

 Return TRUE if a single bracket has been marked (probably by us)

*/

BOOL
FindMarkedBracket(config)

struct EditConfig *config;
{
    if (config->Marker == BLOCKMODE_CHAR) {

        if (config->BlockStartY == config->BlockEndY) {

            if (config->BlockStartX == config->BlockEndX) {

                UWORD pos = config->BlockStartX;

                if (config->BlockStartY == config->Line) {

                    if (pos < config->CurrentLen)

                        return(IsABracket(config->CurrentBuffer[pos]));
                }
                else {

                    struct LineNode *lineNode = config->TextNodes + config->BlockStartY;

                    if (pos < lineNode->Len)

                        return(IsABracket(lineNode->Text[config->BlockStartX]));
                }
            }
        }
    }

    return(FALSE);
}

/* -------------------------------- IsABracket ---------------------------------

 Check whether character is a bracket

*/

BOOL
IsABracket(letter)

UWORD letter;
{
    switch (letter) {

        case '(':
        case ')':
        case '{':
        case '}':
        case '[':
        case ']':

            return(TRUE);

        default:

            return(FALSE);
    }
}

///
