/* -----------------------------------------------------------------------------

 1999 Dietmar Eilert. All Rights Reserved.

 Dietmar Eilert
 Mies-v-d-Rohe-Str.31, 52074 Aachen, Germany
 Phone: +49-(0)179-5987061 German/English
 E-Mail: Dietmar.Eilert@post.rwth-aachen.de
 E-Mail: dietmar_eilert@yahoo.de (alternative address)
 WWW:    http://members.tripod.com/golded
 Mirror: http://members.xoom.com/golded

 Syntax parser example. This example highlights C comments and everything after
 "//" (C++ comments) and uses a syntax chache to be as fast as possible (ie. each
 line is parsed only once).

 DICE:

 DMAKE

 -------------------------------------------------------------------------------

*/

#include "defs.h"

/// "definitions"

// we extend the ParserHandle structure for our private data

struct MyParserHandle {

    struct ParserHandle   ParserHandle;              // embedded parser handle
    struct SyntaxChunk   *SyntaxStack;               // parser output
    struct EditConfig    *EditConfig;                // buffer data
    struct RefreshRequest RefreshRequest;            // display refresh request
};

// empty stack flag ("line has been parsed and contains no syntax elements")

#define EMPTY_STACK ((struct SyntaxInfo *)~0)

// syntax info structures are attached to lines and used to cache syntax information and information about comments

struct SyntaxInfo {

    UWORD                 Flags;                     // comment flags (see below)
    struct SyntaxChunk    SyntaxChunk;               // syntax chunks
};

// define basis size of struct SyntaxInfo (without array of chunks)

#define SIZEOF_STRUCT_SYNTAXINFO sizeof(UWORD)

// these flag bits describe the comment status of each line

#define COMMENT_NONE      (1L<<0)                    // no comment in this line
#define COMMENT_SIMPLE    (1L<<1)                    // there is a comment in this line
#define COMMENT_START     (1L<<2)                    // line is start of multi-line comment
#define COMMENT_BODY      (1L<<3)                    // line is body  of multi-line comment
#define COMMENT_END       (1L<<4)                    // line is end   of multi-line comment

// macros to access SyntaxInfo members

#define GET_FLAGS(node)    (((struct SyntaxInfo *)(node)->UserData)->Flags)
#define GET_CHUNK(node)   (&((struct SyntaxInfo *)(node)->UserData)->SyntaxChunk)

// syntax elements understood by parser

enum SyntaxLevels {

    SYNTAX_STANDARD,
    SYNTAX_COMMENT,
    SYNTAX_COMMENTCPP,
    SYNTAX_MAXIMUM
};

// names of built-in syntax levels

UBYTE *LevelNames[] = { "Standard text", "Comment", "Comment (C++)", NULL };

///
/// "prototypes"

// library functions

Prototype LibCall struct ParserData     *MountScanner(void);
Prototype LibCall struct ParserHandle   *StartScanner(__A0 struct GlobalConfig *, __A1 struct EditConfig *, __D0 struct SyntaxChunk *, __D1 struct SyntaxSetup *);
Prototype LibCall ULONG                  CloseScanner(__A0 struct ParserHandle *);
Prototype LibCall void                   FlushScanner(__A0 struct ParserHandle *);
Prototype LibCall void                   SetupScanner(__A0 struct GlobalConfig *);
Prototype LibCall struct RefreshRequest *BriefScanner(__A0 struct ParserHandle *, __A1 struct ScannerNotify *);
Prototype LibCall struct SyntaxChunk    *ParseLine   (__A0 struct ParserHandle *, __A1 struct LineNode *, __D0 ULONG);
Prototype LibCall void                   UnparseLines(__A0 struct LineNode *, __D0 ULONG);
Prototype LibCall void                   ParseSection(__A0 struct ParserHandle *, __A1 struct LineNode *, __D0 ULONG);

// private functions

Prototype BOOL                           BadSuccessor    (ULONG, ULONG);
Prototype void                           ParseContext    (struct ParserHandle *, struct LineNode *, ULONG);
Prototype struct SyntaxInfo             *ParseContextLine(struct ParserHandle *, struct LineNode *, ULONG);
Prototype struct SyntaxInfo             *DupInfo         (struct SyntaxInfo *);
Prototype struct RefreshRequest         *VerifyContext   (struct ParserHandle *, ULONG, ULONG);
Prototype struct LineNode               *FindContextStart(struct ParserHandle *, struct LineNode *);

///
/// "library functions"

/* ------------------------------- MountScanner --------------------------------

 Called by the editor before first usage of a scanner. Return a description of
 our abilities.

*/

LibCall struct ParserData *
MountScanner()
{
    static UBYTE version[] = "$VER: Comments (context) 3.0 (" __COMMODORE_DATE__ ")";

    static struct ParserData parserData;

    static UBYTE *example[] = {

        "/* Simple ANSI C example */",
        "                           ",
        "#include <defs.h>          ",
        "                           ",
        "int                        ",
        "main(int argc, char **argv)",
        "{                          ",
        "   // anybody at home ?    ",
        "                           ",
        "   puts(\"Hello world !\");",
        "}                          ",

        NULL
    };

    parserData.pd_Release    = SCANLIBVERSION;
    parserData.pd_Version    = 1;
    parserData.pd_Serial     = 0;
    parserData.pd_Info       = "Comments (context) 3.0";
    parserData.pd_Example    = example;
    parserData.pd_Properties = 0;

    // signal editor that we cache syntax information and parse multiple lines (contexts)

    parserData.pd_Flags = SCPRF_SYNTAXCACHE | SCPRF_CONTEXT;

    return(&parserData);
}


/* ------------------------------- StartScanner --------------------------------

 Called by the editor after a new text buffer has been created. We return a
 parser handle (extended with our private data).

*/

LibCall struct ParserHandle *
StartScanner(__A0 struct GlobalConfig *globalConfigPtr, __A1 struct EditConfig *editConfigPtr, __D0 struct SyntaxChunk *syntaxStack, __D1 struct SyntaxSetup *syntaxSetup)
{
    struct MyParserHandle *handle;

    if (handle = (struct MyParserHandle *)AllocVec(sizeof(struct MyParserHandle), MEMF_PUBLIC | MEMF_CLEAR)) {

        handle->ParserHandle.ph_Levels   = SYNTAX_MAXIMUM;
        handle->ParserHandle.ph_Names    = LevelNames;
        handle->ParserHandle.ph_ColorsFG = NULL;
        handle->ParserHandle.ph_ColorsBG = NULL;
        handle->SyntaxStack              = syntaxStack;
        handle->EditConfig               = editConfigPtr;
    }

    return((struct ParserHandle *)handle);
}


/* ------------------------------- CloseScanner --------------------------------

 Called by the editor if a text buffer is about to be closed. Deallocate buffer
 specific 'global' data.

*/

LibCall ULONG
CloseScanner(__A0 struct ParserHandle *parserHandle)
{
    if (parserHandle)

        FreeVec(parserHandle);

    return(0);
}


/* ------------------------------- FlushScanner --------------------------------

 Called by the editor in low memory situations: we are supposed to free as much
 memory as possible. We free cached syntax information.

*/

LibCall void
FlushScanner(__A0 struct ParserHandle *parserHandle)
{
    struct EditConfig *config = ((struct MyParserHandle *)parserHandle)->EditConfig;

    if (config->TextNodes && config->Lines)

        UnparseLines(config->TextNodes, config->Lines);
}


/* ------------------------------- SetupScanner --------------------------------

 Called by the editor if the user wants to change the scanner's configuration.
 We do not support user configuration (parserData.pd_Flags: SCPRF_CONFIGWIN flag
 unset).

*/

LibCall void
SetupScanner(__A0 struct GlobalConfig *globalConfigPtr)
{
    ;
}


/* ------------------------------- BriefScanner --------------------------------

 Called to notify a context scanner if lines have been added, deleted or
 modified. We are supposed to return an additional display request or NULL
 (the editor will refresh the damaged region only if we return NULL). Damaged
 lines have already been unparsed by the editor. This is what we have to do:

 a) lines have been deleted or modified

    Check whether the syntax scheme of the remaining lines is still valid (it
    will be invalid if say a comment start has been deleted but the comment
    end has not been deleted). Reparse the lines and return a refresh request
    if not.

 b) lines have been inserted

    Check whether the syntax scheme of the new lines affects the existing lines.
    Reparse all lines starting at the insertion point and return a refresh
    request if not.

 c) lines have been (un)folded

    This scanner assumes that folding doesn't affect syntax highlighting for the
    sake of simplicity. This isn't necessarily a valid assumption: we will run
    into touble if the user folds parts of a comment only (e.g. the first line).

*/

LibCall struct RefreshRequest *
BriefScanner(__A0 struct ParserHandle *parserHandle, __A1 struct ScannerNotify *notify)
{
    switch (notify->sn_Class) {

        case SCANNER_NOTIFY_MODIFIED:

            {
                ULONG line = notify->sn_Line;

                if (notify->sn_Removed > notify->sn_Lines) {

                    // lines have been deleted

                    return(VerifyContext(parserHandle, line, 0));
                }
                else if (notify->sn_Lines) {

                    // lines have been modified or inserted

                    return(VerifyContext(parserHandle, line, notify->sn_Lines));
                }
                else
                    return(NULL);
            }

            break;

        default:

            return(NULL);
    }
}


/* --------------------------------- ParseLine ---------------------------------

 Parse a line, build a syntax description

*/

LibCall struct SyntaxChunk *
ParseLine(__A0 struct ParserHandle *parserHandle, __A1 struct LineNode *lineNode, __D0 ULONG line)
{
    if (IS_FOLD(lineNode)) {

        return(NULL);
    }
    else if (lineNode->Len) {

        if (lineNode->UserData == EMPTY_STACK) {

            // line parsed already

            return(NULL);
        }
        else if (lineNode->UserData) {

            // line parsed already

            return(GET_CHUNK(lineNode));
        }
        else {

            ParseSection(parserHandle, lineNode, 1);

            if (lineNode->UserData == EMPTY_STACK)
                return(NULL);
            else
                return(GET_CHUNK(lineNode));
        }
    }
    else
        return(NULL);
}


/* -------------------------------- UnparseLines -------------------------------

 Called by the editor if lines are to be deleted. We are supposed to free
 private data attached to the lines (cached syntax information).

*/

LibCall void
UnparseLines(__A0  struct LineNode *lineNode, __D0 ULONG lines)
{
    while (lines--) {

        // free syntax cache

        if (lineNode->UserData) {

            if (lineNode->UserData != (APTR)EMPTY_STACK)

                FreeVec((APTR)lineNode->UserData);

            lineNode->UserData = NULL;
        }

        // free folded subblock

        if (IS_FOLD(lineNode)) {

            struct Fold *fold = (struct Fold *)lineNode->SpecialInfo;

            UnparseLines(fold->TextNodes, fold->Lines);
        }

        ++lineNode;
    }
}


/* -------------------------------- ParseSection -------------------------------

 Called by the editor if lines are to be displayed. The scanner is encouraged to
 preparse the lines.

*/

LibCall void
ParseSection(__A0 struct ParserHandle *parserHandle, __A1 struct LineNode *lineNode, __D0 ULONG lines)
{
    struct LineNode *firstNode;

    if (firstNode = FindContextStart(parserHandle, lineNode))

        ParseContext(parserHandle, firstNode, lines + (lineNode - firstNode));
}

///
/// "private"

/*  ----------------------------- FindContextStart ------------------------------

 Search backwards until a parsed line or a "context free line" is found, ie. a
 line which can't possibly be affected by the lines before it as far as syntax
 parsing is concerned. The first line of a document is a context free line. We
 assume that the user doesn't use folds inside comments, so fold headers can be
 considered context-free lines, too (all comments end before the fold). The node
 returned by this function may be used as starting point for parsing a context.

*/

struct LineNode *
FindContextStart(parserHandle, lineNode)

struct ParserHandle *parserHandle;
struct LineNode     *lineNode;
{
    ULONG line = lineNode - ((struct MyParserHandle *)parserHandle)->EditConfig->TextNodes;

    while (line--) {

        // found a fold or a parsed line ?

        if (lineNode->UserData || IS_FOLD(lineNode))

            break;

        --lineNode;
    }

    return(lineNode);
}


/* ------------------------------- ParseContext --------------------------------

 Preparse lines. The first line is expected to be either a context free line
 (start of file or a fold header) or to be a parsed line.

*/

void
ParseContext(parserHandle, lineNode, lines)

struct ParserHandle *parserHandle;
struct LineNode     *lineNode;
ULONG                lines;
{
    ULONG mode;

    for (mode = COMMENT_NONE; lines--; ++lineNode) {

        if (IS_FOLD(lineNode)) {

            mode = COMMENT_NONE;
        }
        else {

            if (lineNode->UserData == EMPTY_STACK) {

                mode = COMMENT_NONE;
            }
            else if (lineNode->UserData) {

                mode = GET_FLAGS(lineNode);
            }
            else {

                struct SyntaxInfo *syntaxInfo = ParseContextLine(parserHandle, lineNode, mode);

                if (syntaxInfo == EMPTY_STACK) {

                    lineNode->UserData = EMPTY_STACK;

                    mode = COMMENT_NONE;
                }
                else if (syntaxInfo) {

                    lineNode->UserData = (APTR)DupInfo(syntaxInfo);

                    mode = syntaxInfo->Flags;
                }
                else
                    mode = COMMENT_NONE;
            }
        }
    }
}


/* -------------------------------- ParseContextLine ---------------------------

 Parse a string, build a syntax description. Return EMPTY_STACK if there is
 nothing to highlight. The status of the last line is passed in by the caller

*/

struct SyntaxInfo *
ParseContextLine(parserHandle, lineNode, last)

struct ParserHandle *parserHandle;
struct LineNode     *lineNode;
ULONG                last;
{
    UWORD len;

    if (len = lineNode->Len) {

        struct SyntaxInfo  *syntaxInfo;
        struct SyntaxChunk *syntaxStack;
        UBYTE              *text;
        UWORD               indent;
        UWORD               element;
        UWORD               commentstart;
        UWORD               status;
        BOOL                instring;
        BOOL                incomment;
        BOOL                localcomment;

        syntaxInfo = (struct SyntaxInfo *)((struct MyParserHandle *)parserHandle)->SyntaxStack;

        text = lineNode->Text;

        // leading spaces and tabs have to be ignored (we ignore all non-ascii characters)

        for (indent = 0; len && (*text <= 32); ++indent) {

            ++text;
            --len;
        }

        // trailing white space has to be ignored

        while (len && (text[len - 1] <= 32))

            --len;

        // initial comment status depends on status of previous line

        if (last & (COMMENT_START | COMMENT_BODY)) {

            // this line is part of a comment started earlier

            status       = COMMENT_BODY;
            commentstart = indent;
            incomment    = TRUE;
        }
        else {

            status       = COMMENT_NONE;
            commentstart = FALSE;
            incomment    = FALSE;
        }

        instring     = FALSE;
        localcomment = FALSE;

        syntaxStack = &syntaxInfo->SyntaxChunk;

        for (element = 0; len; ++text, ++indent, --len) {

            if (incomment) {

                // end of comment detected ?

                if ((*text == '*') && (text[1] == '/') && (len > 1)) {

                    // end of a local comment (comment starting in the same line) ?

                    if (localcomment) {

                        status |= COMMENT_SIMPLE;
                    }
                    else {

                        status |=  COMMENT_END;
                        status &= ~COMMENT_BODY;
                    }

                    syntaxStack[element].sc_Level = SYNTAX_COMMENT;
                    syntaxStack[element].sc_Start = commentstart;
                    syntaxStack[element].sc_End   = indent + 1;

                    ++element;

                    incomment = FALSE;

                    ++text;
                    ++indent;
                    --len;
                }
            }
            else if ((*text == '\"') && ((*(text - 1) != 92) || (*(text - 2) == 92))) {

                instring = !instring;
            }
            else if (instring == FALSE) {

                if ((*text == '/') && (text[1] == '/') && (len > 1)) {

                    // C++ comment detected

                    syntaxStack[element].sc_Level = SYNTAX_COMMENTCPP;
                    syntaxStack[element].sc_Start = indent;
                    syntaxStack[element].sc_End   = indent + len - 1;

                    ++element;

                    // terminate parsing (rest of line is a comment)

                    break;
                }
                else if ((*text == '/') && (text[1] == '*') && (len > 1)) {

                    // start of comment detected detected in this line

                    commentstart = indent;
                    incomment    = TRUE;
                    localcomment = TRUE;

                    ++text;
                    ++indent;
                    --len;
                }
            }
        }

        // end of line reached but comment not closed ?

        if (incomment) {

            // comment starting in this line ?

            if (localcomment)
                status |= COMMENT_START;
            else
                status |= COMMENT_BODY;

            syntaxStack[element].sc_Level = SYNTAX_COMMENT;
            syntaxStack[element].sc_Start = commentstart;
            syntaxStack[element].sc_End   = indent - 1;

            ++element;
        }

        if (element) {

            // terminate syntax stack

            syntaxStack[element].sc_Start = FALSE;
            syntaxStack[element].sc_End   = FALSE;
            syntaxStack[element].sc_Level = FALSE;

            syntaxInfo->Flags = status;

            return(syntaxInfo);
        }
        else
            return(EMPTY_STACK);
    }
    else if (last) {

        if (last & (COMMENT_START | COMMENT_BODY)) {

            struct SyntaxInfo *syntaxInfo = (struct SyntaxInfo *)((struct MyParserHandle *)parserHandle)->SyntaxStack;

            // line is empty but part of a multi-line comment

            syntaxInfo->SyntaxChunk.sc_Start = FALSE;
            syntaxInfo->SyntaxChunk.sc_End   = FALSE;
            syntaxInfo->SyntaxChunk.sc_Level = FALSE;

            syntaxInfo->Flags = COMMENT_BODY;

            return(syntaxInfo);
        }
        else
            return(EMPTY_STACK);
    }
    else
        return(EMPTY_STACK);
}


/* ------------------------------- VerifyContext -------------------------------

 Ensure that syntax information of a specified range of lines ("damage region")
 is valid and consistent with the existing text. Return a refresh request if
 not. The damage area is expected to have been unparsed already. This is what we
 have to do: we preparse existing lines before the damage area if belonging to
 the syntax context of the damage area (ie. all lines affecting highlighting of
 the first line in the damage area). The damage area is parsed, too. Parsed
 lines after the damage area are reparsed if highlighting is found to be
 inconsistent with the line(s) before. Reparsing continues until the end of the
 file respectively until no more inconsistencies are found.

*/

struct RefreshRequest *
VerifyContext(parserHandle, line, lines)

struct ParserHandle *parserHandle;
ULONG                line;
ULONG                lines;
{
    struct EditConfig *config;
    struct LineNode   *lineNode;
    struct LineNode   *lastNode;
    struct SyntaxInfo *syntaxInfo;
    ULONG              refreshStart;
    ULONG              refresh;
    ULONG              last;
    ULONG              new;

    config = ((struct MyParserHandle *)parserHandle)->EditConfig;

    lineNode = config->TextNodes + line;
    lastNode = config->TextNodes + line + (lines - 1);

    // preparse from context start until end of damage area

    ParseSection(parserHandle, lineNode, lines);

    // get syntax flags of last line in damage area

    if ((lastNode < config->TextNodes) || IS_FOLD(lastNode) || (lastNode->UserData == EMPTY_STACK) || (lastNode->UserData == NULL))
        last = COMMENT_NONE;
    else
        last = GET_FLAGS(lastNode);

    // continue parsing until no more inconsistencies are found (until context end)

    refreshStart = (line += lines);
    refresh      = FALSE;

    for (lineNode = lastNode + 1; line < config->Lines; ++line, ++lineNode, ++refresh) {

        // folds terminate parsing (context end)

        if (IS_FOLD(lineNode)) {

            break;
        }
        else if (lineNode->UserData == NULL) {

            // line not yet parsed

            syntaxInfo = ParseContextLine(parserHandle, lineNode, last);

            if (syntaxInfo == EMPTY_STACK) {

                lineNode->UserData = EMPTY_STACK;
            }
            else if (syntaxInfo) {

                lineNode->UserData = DupInfo(syntaxInfo);
            }
        }
        else {

            // check whether highlighting of this line is consistent with previous line

            new = (lineNode->UserData == EMPTY_STACK) ? COMMENT_NONE : GET_FLAGS(lineNode);

            if (BadSuccessor(last, new)) {

                if (lineNode->UserData != EMPTY_STACK)

                    FreeVec(lineNode->UserData);

                if (syntaxInfo = ParseContextLine(parserHandle, lineNode, last)) {

                    if (syntaxInfo == EMPTY_STACK)
                        lineNode->UserData = EMPTY_STACK;
                    else
                        lineNode->UserData = DupInfo(syntaxInfo);
                }
                else
                    lineNode->UserData = EMPTY_STACK;
            }
            else
                break;
        }

        if ((lineNode->UserData == EMPTY_STACK) || (lineNode->UserData == NULL) || IS_FOLD(lineNode))
            last = COMMENT_NONE;
        else
            last = GET_FLAGS(lineNode);
    }

    if (refresh) {

        struct RefreshRequest *refreshRequest = &((struct MyParserHandle *)parserHandle)->RefreshRequest;

        refreshRequest->rr_Line  = refreshStart;
        refreshRequest->rr_Lines = refresh;

        return(refreshRequest);
    }
    else
        return(NULL);
}


/* ------------------------------- BadSuccessor --------------------------------

 Return TRUE if syntax information of two adjacent lines is inconsistent.

*/

BOOL
BadSuccessor(pred, succ)

ULONG pred, succ;
{
    if (succ & (COMMENT_BODY | COMMENT_END)) {

        // comment body/end without start ?

        if ((pred & (COMMENT_START | COMMENT_BODY)) == FALSE)

            return(TRUE);
    }
    else if (pred & COMMENT_START) {

        // comment start without body/end ?

        if ((succ & (COMMENT_BODY | COMMENT_END)) == FALSE)

            return(TRUE);

        // comment inside comment ?

        if (succ & COMMENT_START)

            return(TRUE);
    }
    else if (pred & COMMENT_BODY) {

        // body line without end ?

        if ((succ & (COMMENT_BODY | COMMENT_END)) == FALSE)

            return(TRUE);
    }

    // C++ comment inside ANSI comment ?

    if (pred & (COMMENT_START | COMMENT_BODY))

        if (pred & COMMENT_SIMPLE)

            return(TRUE);

    return(FALSE);
}


/* --------------------------------- DupInfo -----------------------------------

 Duplicate syntax info (to be FreeVec'ed)

*/

struct SyntaxInfo *
DupInfo(syntaxInfo)

struct SyntaxInfo *syntaxInfo;
{
    struct SyntaxChunk *chunk;
    struct SyntaxInfo  *info;
    ULONG               size;
    UWORD               elements;

    // determine stack size

    for (elements = 0, chunk = &syntaxInfo->SyntaxChunk; chunk->sc_Level; ++chunk)

        ++elements;

    // create copy of syntax stack (to be attached to a text line by the caller)

    size = SIZEOF_STRUCT_SYNTAXINFO + (++elements) * sizeof(struct SyntaxChunk);

    if (info = (struct SyntaxInfo  *)AllocVec(size, MEMF_PUBLIC)) {

        movmem(syntaxInfo, info, size);

        return(info);
    }
    else
        return(NULL);
}

///
